local SHOT_BELOW_Y, SHOT_STRAIGHT_Y = 25, -5 
local DAMAGE_FROM_HIT, DEAD_UP_MOVMENT = 10, 100
local HIT_ANI, MOVE_ANI, IDLE_ANI, KILLFALL, STRIKE1, STRIKE2, STRIKE3 = "LEVEL_CULTIST_HITHIGH", "LEVEL_CULTIST_ADVANCE", "LEVEL_CULTIST_IDLE", 
"LEVEL_CULTIST_KILLFALL", "LEVEL_CULTIST_STRIKE1", "LEVEL_CULTIST_STRIKE2", "LEVEL_CULTIST_STRIKE3"
local BASE, MOVE, MOVE_APROACH, HITBACK, SHOOTING, RUNNING, IDLE, IDLE_AGRESSIVE, HITTING, AFTER_HITTING, DEAD = 0, 1, 2, 41, 11, 3, 4, 5, 12, 13, 404
local SHOOTING_RANGE, ATTACK_APROACH_RANGE, ATTACKING_RANGE = 550, 400, 60
local IDLE_DURATION, IGNORE_CLAW_TIME, MINIMAL_MAN_UP_TIME, CAMELEON_SHOT_COLDOWN = 1540, 2500, 1500, 3000
local HITBACK_DURATION, SHOOTING_DURATION = 200, 320
local SOUND_DEATH = "LEVEL_CULTIST_DAMAGE"
local CLAW_SOUNDS_MEELE = {"CLAW_1001","CLAW_1007A","CLAW_1004A"}
local CLAW_SOUNDS_RANGED = {"CLAW_1003", "CLAW_1009", "CLAW_1110064"}
local SHOT_RADIUS = 10

-- Hit States
local HitStates = {
	NotStarted = 1,
	PreStrike = 2,
	Strike = 3,
	AfterStrike = 4
}

local function createChameleonShot(self)
	self.CameleonThrowsLeft = self.CameleonThrowsLeft - 1
	self.StopTime = GetTicks()
	
	local yShotChange
	
	if(self.ShotAbove) then
		yShotChange = SHOT_STRAIGHT_Y
	else
		yShotChange = SHOT_BELOW_Y
	end
	
	local claw = GetClaw()

	self.Projectile = CreateObject {x=self.X, y= self.Y+yShotChange, z=self.Z, image = "GAME_MAGICCLAW", logic="CustomLogic", name="CameleonShot", Damage = 10, SpeedX = 3, Direction = self.Direction}
	PlaySound("LEVEL_CHAMELEON_CHAMELEONJUMP2")
end

local function changeAttacktRect(self, direction)
	if(direction == -1) then
		self.AttackRect = {-25,-40,15,40}
	else
		self.AttackRect = {-15,-40,32,40}
	end
end

local function changeDirection(self, direction)
	self.Direction = direction
	changeAttacktRect(self, direction)
	if(direction == -1) then
		self.DrawFlags.Mirror = false
		self.HitRect = {-25,-40,15,40}
	else
		self.DrawFlags.Mirror = true
		
		self.HitRect = {-15,-40,32,40}
	end
end

local function playClawSoundAfterKill(weaponFlag)
	math.randomseed(os.time())
	local randomNumber = math.random(1, 6)
	if(randomNumber > 3) then 
		do return end
	end
	if weaponFlag == 2 then
		ClawSound(CLAW_SOUNDS_MEELE[randomNumber])
	end
	if weaponFlag == 32 then
		ClawSound(CLAW_SOUNDS_RANGED[randomNumber])
	end

end


function init(self)
	
	self.UpMovment = DEAD_UP_MOVMENT
	self.StopTime = GetTicks()
	self.CameleonThrowsLeft = 2
	self.StartedShot = false
	self.StateForcedDuration = 0
	self.StartOfForcedState = GetTicks()
	self.YShotChange = SHOT_STRAIGHT_Y
	self.CreatedUnit = false
	self.Health = 30
	self.Z = 4010
	self.HitTick = GetTicks()
	self.ProjectileHitTick = GetTicks()
	self.AttackTick = GetTicks()
	self.IdleStart = GetTicks()
	self.LastHitTick = GetTicks()
	self.Damage = 5
	self.MaxSpeedX  = 2
	self.DrawFlags.Mirror = true
	self.Speed = 1
	self.StartingAproaching = GetTicks()
	self.HitState = HitStates.NotStarted
	self.StartedMannigUp = GetTicks()
	self.StartIgnoreClaw = GetTicks()
	self.Treasures = {self.UserRect1.Bottom,self.UserRect1.Top,self.UserRect1.Left,self.UserRect1.Right}
	self.SpeedX = 2
	self.Gravity = -7
	changeDirection(self, 1)
end
	
function main(self)
	local claw = GetClaw()

	if(self.State == DEAD) then
		self.Gravity, self.X, self.Y = self.Gravity + 0.25, self.X + self.SpeedX*self.Direction, self.Y + self.Gravity
		if(self.Y - claw.Y > 450 ) then
			PlaySound("GAME_SPLASH")
			self:Destroy()
		end
		do return end
	end

	self.PhysicsType = 8
	self:AnimationStep()
	
	self:AlignToGround()
	if self.State == BASE then
		self.StateForcedDuration = 0
		self.CurrentSpeedX  = 2
		self.MoveRect = {-20,-40,22,60}
		self.HitTypeFlags = 50335794
		self.ObjectTypeFlags = 4
		self:SetImage("LEVEL_CULTIST")
		changeAttacktRect(self, self.direction)
	end
	

	if(self.State == HITBACK) then
		if GetTicks()-self.StartOfForcedState>self.StateForcedDuration then
			self.State = BASE
		else
			do return end
		end
	end
	
	if(self.State == SHOOTING) then
		if GetTicks()-self.StartOfForcedState>self.StateForcedDuration then
			createChameleonShot(self)
			self.StateForcedDuration = 200
			self.State = AFTER_HITTING
		end
		do return end
	end
	
	if(self.State == HITTING) then
		if(self.HitState == HitStates.NotStarted) then
			if GetTicks()-self.LastHitTick> 300 then
				self.StartOfForcedState = GetTicks()
				self.StateForcedDuration = 245
				self.HitState =  HitStates.PreStrike
				self:SetAnimation(STRIKE1)
			end
		end
		
		if(self.HitState == HitStates.PreStrike) then
			if GetTicks()-self.StartOfForcedState>self.StateForcedDuration then
				self.LastHitTick = GetTicks()
				if self.Direction == -1 then
					self.AttackRect = { -60,-30,-20,10}
				else
					self.AttackRect = { 20,-30,60,10}
				end
				self.StartOfForcedState = GetTicks()
				self.StateForcedDuration = 110
				self.HitState =  HitStates.Strike
			end
		end
		
		if(self.HitState == HitStates.Strike) then
			if GetTicks()-self.StartOfForcedState>self.StateForcedDuration then
				self.StartOfForcedState = GetTicks()
				self.StateForcedDuration = 160
				self.State = AFTER_HITTING
			end
		end
		do return end
	end
	
	if(self.State == AFTER_HITTING) then
		if GetTicks()-self.StartOfForcedState>self.StateForcedDuration then
			self.State = BASE
		end
		do return end
	end
	
	

	local claw = GetClaw()
	

	local clawMaximalY = claw.Y + claw.HitRect.Bottom
	local clawMinimalY = claw.Y + claw.HitRect.Top

	
	local isClawInYRange = false
	
	if (clawMinimalY <= self.Y + SHOT_STRAIGHT_Y + SHOT_RADIUS and clawMaximalY >= self.Y + SHOT_STRAIGHT_Y - SHOT_RADIUS) then
		isClawInYRange = true
		self.ShotAbove = true
	elseif (clawMaximalY >= self.Y + SHOT_BELOW_Y - SHOT_RADIUS and clawMinimalY <= self.Y + SHOT_BELOW_Y + SHOT_RADIUS)  then
		isClawInYRange = true
		self.ShotAbove = false
	end
	
	local canShoot = GetTime() - self.StopTime > CAMELEON_SHOT_COLDOWN and self.CameleonThrowsLeft > 0
	local isClawInShootingRange = math.abs(claw.X - self.X) < SHOOTING_RANGE
	
	if isClawInYRange and canShoot and isClawInShootingRange then
		if(self.X > claw.X) then
			changeDirection(self, -1)
		else
			changeDirection(self, 1)
		end
		self.State = SHOOTING
		self.StartOfForcedState = GetTicks()
		self.StateForcedDuration = SHOOTING_DURATION
		
		if(self.ShotAbove) then
			self:SetAnimation(STRIKE2)
		else
			self:SetAnimation(STRIKE3)
		end
		do return end
	end
	
	if isClawInYRange and self.CameleonThrowsLeft > 0 and isClawInShootingRange then
		if(GetTicks()-self.StartedMannigUp>MINIMAL_MAN_UP_TIME) then
		
			local hasSpaceToRunAway =  (self.X > claw.X and self.XMax - self.X > 200)  or (self.X < claw.X and self.X - self.XMin > 200) 
		
			if(hasSpaceToRunAway) then

				if(self.X > claw.X) then
					changeDirection(self, 1)
				else 
					changeDirection(self, -1)
				end
				
				if(self.State ~= RUNNING) then
					self.State = RUNNING
					
					self:SetAnimation(MOVE_ANI)
				end
				self.X = self.X + self.MaxSpeedX * self.Direction
				do return end
			else
				self.StartedMannigUp = GetTicks()
			end
		end
	end
	
	local isClawInAttackRange = math.abs(claw.X - self.X) < ATTACKING_RANGE
	
	if(isClawInYRange) then
		if(isClawInAttackRange) then
			if(self.X > claw.X) then
				changeDirection(self, -1)
			else 
				changeDirection(self, 1)
			end
			self.HitState = HitStates.NotStarted
			self.State = HITTING
			do return end
		end

		local clawOutsideMovmentBoundary = (self.State == IDLE or self.State == IDLE_AGRESSIVE) and ((claw.X > self.X and self.Direction == 1) or (claw.X < self.X and self.Direction == -1))
		if math.abs(claw.X - self.X) < ATTACK_APROACH_RANGE and GetTicks() - self.StartIgnoreClaw > 2500 and not clawOutsideMovmentBoundary then
			if(self.State ~= MOVE_APROACH)then
				self.State = MOVE_APROACH
				self:SetAnimation(MOVE_ANI)
			end
			if(self.X > claw.X) then
				changeDirection(self, -1)				
			else 
				changeDirection(self, 1)
			end
		end
	end
	
	if self.State == IDLE then
		if GetTime() - self.IdleStart > IDLE_DURATION then	
			self.StartIgnoreClaw = GetTicks()
			changeDirection(self, self.Direction * -1)
			self.State = BASE
		end
		do return end
	end
	
	if self.State == IDLE_AGRESSIVE then
		if GetTime() - self.IdleStart > IDLE_DURATION then	
			changeDirection(self, self.Direction * -1)
			self.StartIgnoreClaw = GetTicks()
			self.State = BASE
		end
		do return end
	end
	
	if self.State ~=  MOVE and self.State ~= MOVE_APROACH then
		self.State = MOVE
		self:SetAnimation(MOVE_ANI)
	end
	
	local speed = self.Speed
	if(self.State == MOVE_APROACH) then speed = self.MaxSpeedX end
	if self.Direction == 1 then
		self.X = self.X + speed
		
		if self.X + speed > self.XMax then	
			if(self.State == MOVE_APROACH) then 
				self.State = IDLE_AGRESSIVE
			else
				self.State = IDLE
			end
			
			self.State = IDLE
			self.IdleStart = GetTicks()
			self:SetAnimation(IDLE_ANI)
		end
	else
		self.X = self.X - speed
		
		if self.X - speed < self.XMin then		
			if(self.State == MOVE_APROACH) then 
				self.State = IDLE_AGRESSIVE
			else
				self.State = IDLE
			end
			self.IdleStart = GetTicks()
			self:SetAnimation(IDLE_ANI)
		end
	end
		
	
end

function getHitAni(self)
	self:SetAnimation(HIT_ANI)
end

function setHitState(self, duration)
	self.StartOfForcedState = GetTicks()
	self.State = HITBACK
	self.StateForcedDuration = duration
end

function hit(self)
	local claw = GetClaw()
	local h = self.HitRef

	-- 4096 is for jump switch
	if self.State == DEAD then

		if  h.ObjectTypeFlags~=33554432 and h.ObjectTypeFlags~=32 and  h.ObjectTypeFlags~=2 then
			do return end
		end
	
		if GetTicks()-self.HitTick > 200 then

			PlaySound(SOUND_DEATH)
			local hitburst = CreateObject {x=self.X, y=h.Y+h.AttackRect.Top - 10, z=self.Z+1, logic="CustomLogic", name="HitBurst", image="GAME_ENEMYHIT"}
			if self.X > claw.X then
				self.Direction = 1
			else
				self.Direction = -1
			end
			self.Gravity = -7
			self.HitTick = GetTicks() 
		end
		do return end
	end
	
	
	
	local hitHigh = h.Y+h.AttackRect.Top < self.Y
	


	if h.ObjectTypeFlags==33554432 or h.ObjectTypeFlags==32 or h.ObjectTypeFlags==33554432 then
		if GetTicks()-self.ProjectileHitTick>300 then
			
			local hitburst = CreateObject {x=self.X, y=h.Y+h.AttackRect.Bottom, z=self.Z+1, logic="CustomLogic", name="HitBurst", image="GAME_ENEMYHIT"}
			self.ProjectileHitTick = GetTicks()
			getHitAni(self)
			setHitState(self, HITBACK_DURATION)

			if(h.ObjectTypeFlags==32) then -- shot
				local damage = 10
				if hitHigh then damage = damage + 5 end
				self.Health = self.Health - damage
				h:Destroy()
			end
			
			if(h.ObjectTypeFlags==33554432) then  -- fire/ice/electric swords/magic claw
				self.Health = self.Health - 25
			end
		end
	end
	
	if(h.ObjectTypeFlags==2) then -- hit
		
		if GetTicks()-self.HitTick > 300 then
	
			setHitState(self, HITBACK_DURATION)
			PlaySound("GAME_HIT1")
			getHitAni(self)
			local hitburst = CreateObject {x=self.X, y=h.Y+h.AttackRect.Bottom, z=self.Z+1, logic="CustomLogic", name="HitBurst", image="GAME_ENEMYHIT"}
			self.HitTick = GetTicks()
			local damage = 5
			if claw.I == 283 then damage = damage + 5 end  -- jump hit
			if hitHigh then damage = damage + 1 end -- hit top part
			self.Health = self.Health - damage		
		end	
	end
	
	if(h.ObjectTypeFlags==16777216) then  -- dynamite and powderkeg
		setHitState(self, HITBACK_DURATION)
		getHitAni(self)
		local hitburst = CreateObject {x=self.X, y=h.Y+h.AttackRect.Top, z=self.Z+1, logic="CustomLogic", name="HitBurst", image="GAME_ENEMYHIT"}
		self.Health = self.Health - 15

	end

	if self.Health <= 0 then
		self:SetFrame(50)
		self.AttackRect = {0, 0, 0, 0}
		if self.Treasures[1] ~= 0 then self.bonus2 = CreateGoodie {x=self.X-5, y=self.Y-10, z=self.Z+1, powerup=self.Treasures[1]} end
		if self.Treasures[2] ~= 0 then self.bonus3 = CreateGoodie {x=self.X, y=self.Y-10, z=self.Z+1, powerup=self.Treasures[2]} end
		if self.Treasures[3] ~= 0 then self.bonus4 = CreateGoodie {x=self.X+5, y=self.Y-10, z=self.Z+1, powerup=self.Treasures[3]} end
		if self.Treasures[4] ~= 0 then self.bonus5 = CreateGoodie {x=self.X+5, y=self.Y-10, z=self.Z+1, powerup=self.Treasures[4]} end
		
		
		-- handling variables needed for falling and hitting falling cultist
		if self.X > claw.X then
			self.Direction = 1
		else
			self.Direction = -1
		end
		
		
		playClawSoundAfterKill(h.ObjectTypeFlags)
		PlaySound(SOUND_DEATH)
		self.ObjectTypeFlags = 1
		self.HitRect = {-40,-28, 64,48}
		self.HitTick = GetTicks() 
		self.State = DEAD
		self.UpMovment = DEAD_UP_MOVMENT
	end
end

function attack(self)

	if false then
		self.AttackTick = GetTicks()
		ClawTakeDamage(10)
	end
end

